/* Copyright 2020-2022, Michele Santullo
 * This file is part of wrenpp.
 *
 * Wrenpp is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Wrenpp is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with wrenpp.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "string_bt.hpp"
#include <cstddef>
#include <utility>

namespace wren {
	namespace detail {
		consteval std::size_t find_last (const char* str, std::size_t len, char c) {
			for (std::size_t z = len; z > 0; --z) {
				if (str[z - 1] == c)
					return z - 1;
			}
			return len;
		}

		consteval std::size_t find_token_end (const char* str, std::size_t len) {
			for (std::size_t z = 0; z < len; ++z) {
				const bool is_lower = str[z] >= 'a' and str[z] <= 'z';
				const bool is_upper = str[z] >= 'A' and str[z] <= 'Z';
				const bool is_num = str[z] >= '0' and str[z] <= '9';
				const bool is_special = str[z] == '_' or str[z] == ':';
				if (not (is_lower or is_upper or is_num or is_special))
					return z;
			}
			return len;
		}

		template <typename T>
		consteval auto guess_class_name_impl() {
#if defined(__GNUC__)
			constexpr auto fname = __PRETTY_FUNCTION__;
			constexpr auto fname_len = sizeof(__PRETTY_FUNCTION__) / sizeof(__PRETTY_FUNCTION__[0]) - 1;
			static_assert(fname_len > 0);

#   if defined(__clang__)
			//clang: void figure_out_name() [T = MyTest]
#   else
			//gcc: void figure_out_name() [with T = MyTest]
#   endif
			constexpr auto last_colon = detail::find_last(fname, fname_len, ':');
			constexpr auto last_space = detail::find_last(fname, fname_len, ' ');
			constexpr auto last_whatever = (last_colon == fname_len ? last_space : last_colon);
			constexpr auto from = last_whatever + (last_whatever == fname_len ? 0 : 1);
			constexpr auto len = detail::find_token_end(fname + from, fname_len - from);

			constexpr dhandy::bt::string<len> retval{fname + from, 0};
			return retval;
	//#elif defined(_MSC_VER)
			//void __cdecl guess_class_name<class MyTest>(void)
			//__FUNCSIG__
#else
#   error "Unsupported compiler"
#endif
		}

		template <typename T, std::size_t S=guess_class_name_impl<T>().size()>
		static constinit dhandy::bt::string g_class_names = {guess_class_name_impl<T>()};
	} //namespace detail

	template <typename T>
	const char* guess_class_name() {
		return detail::g_class_names<T>.data();
	}
} //namespace wren
