/* Copyright 2020-2022, Michele Santullo
 * This file is part of wrenpp.
 *
 * Wrenpp is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Wrenpp is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with wrenpp.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

typedef struct WrenHandle WrenHandle;

namespace wren {
	class VM;

	class Handle {
	public:
		Handle() noexcept;
		Handle (const Handle&) = delete;
		Handle (Handle&& other) noexcept;
		Handle (VM* vm, WrenHandle* handle) noexcept :
			m_handle(handle),
			m_vm(vm)
		{
		}

		Handle& operator= (const Handle&) = delete;
		Handle& operator= (Handle&& other) noexcept;

		~Handle() noexcept;
		void release() noexcept;

		operator WrenHandle*() const { return m_handle; }
		operator bool() const { return nullptr != m_handle; }

	private:
		WrenHandle* m_handle;
		VM* m_vm;
	};

	inline Handle::Handle() noexcept :
		m_handle(nullptr),
		m_vm(nullptr)
	{
	}

	inline Handle::Handle (Handle&& other) noexcept :
		m_handle(other.m_handle),
		m_vm(other.m_vm)
	{
		other.m_handle = nullptr;
		other.m_vm = nullptr;
	}

	inline Handle& Handle::operator= (Handle&& other) noexcept {
		{
			auto tmp = other.m_handle;
			other.m_handle = m_handle;
			m_handle = tmp;
		}
		{
			auto tmp = other.m_vm;
			other.m_vm = m_vm;
			m_vm = tmp;
		}
		return *this;
	}
} //namespace wren
