/* Copyright 2020-2022, Michele Santullo
 * This file is part of wrenpp.
 *
 * Wrenpp is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Wrenpp is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with wrenpp.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "wrenpp/def_configuration.hpp"
#include "wrenpp/callback_manager.hpp"
#include "wrenpp/class_manager.hpp"
#include "wrenpp/vm.hpp"
#include <iostream>
#include <algorithm>
#include <cstdlib>

namespace wren {
	namespace {
		template <typename T>
		bool empty(T* message) {
			return not (message and *message);
		}
	} //unnamed namespace

	void DefConfiguration::write_fn (VM*, wren_string_t text) {
		std::cout << text;
	}

	void DefConfiguration::error_fn (VM*, ErrorType type, wren_string_t module, int line, wren_string_t message) {
		using std::empty;

		std::cerr << "Wren error: " << message;
		if (!empty(module)) {
			std::cerr << " in " << module;
			if (line > -1) {
				std::cerr << ':' << line;
			}
		}
		std::cerr << '\n';
	}

	void* DefConfiguration::reallocate_fn (void* ptr, std::size_t size) {
		if (0 == size) {
			std::free(ptr);
			return nullptr;
		}

		return std::realloc(ptr, size);
	}

	foreign_method_t DefConfiguration::foreign_method_fn (VM* vm, wren_string_t module_name, wren_string_t class_name, bool is_static, wren_string_t signature) {
		return vm->callback_manager().callback(is_static, module_name, class_name, signature);
	}

	foreign_class_t DefConfiguration::foreign_class_fn (wren::VM* vm, std::string_view module_name, std::string_view class_name) {
		return vm->class_manager().make_class(module_name, class_name);
	}
} //namespace wren
